﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.UI;
using MeshVR;


namespace JayJayWon
{

    public abstract class BaseMVRS
    {
        protected static MVRScript MVRS;

        public BaseMVRS(MVRScript mvrs)
        {
            MVRS = mvrs;
        }
    }
   

    public class ButtonRowUIElement : BaseMVRS
    {
        protected UIDynamic _spacerLeft { get; set; }
        protected UIDynamic _spacerRight { get; set; }

        private int spacerHeight ;
        private int buttonPadding = 2;

        private Dictionary<int, UIDynamicButton> buttonDict;
        public ButtonRowUIElement(MVRScript mvrs, int height = 50) : base(mvrs)
        {
            spacerHeight = height;
            buttonDict = new Dictionary<int, UIDynamicButton>();
        }

        private UIDynamicButton CreateUIButton()
        {
            Transform buttonTransform = UnityEngine.Object.Instantiate(MVRS.manager.configurableButtonPrefab.transform, _spacerLeft.gameObject.transform, false);
            UIDynamicButton selectTargetButton = buttonTransform.GetComponent<UIDynamicButton>();
            buttonTransform.SetParent(_spacerLeft.gameObject.transform, false);
            return (selectTargetButton);
        }

        private void ClearAllButtonColors()
        {
            foreach (KeyValuePair<int, UIDynamicButton> kvp in buttonDict)
            {
                kvp.Value.buttonColor = new Color(0.85f, 0.85f, 0.85f, 1f);
            }
        }

        public void AddButtonToRow(string label, int buttonRef, Action<int> callbackMethod, int selectedButtonRef)
        {
            UIDynamicButton newButton = CreateUIButton();
            newButton.label = label;

            if (selectedButtonRef==buttonRef) newButton.buttonColor = Color.cyan;

            newButton.button.onClick.AddListener(delegate
            {
                try
                {
                    callbackMethod.Invoke(buttonRef);
                }
                catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
            });

            buttonDict.Add(buttonRef, newButton);
        }

        public void SetButtonColorActive(int buttonRef)
        {
            ClearAllButtonColors();
            buttonDict[buttonRef].buttonColor = Color.cyan;
        }

        public void ClearUI()
        {
            if (_spacerLeft != null) MVRS.RemoveSpacer(_spacerLeft);
            if (_spacerRight != null) MVRS.RemoveSpacer(_spacerRight);
            buttonDict.Clear();
        }

        public void RefreshUI(int buttonCount)
        {
            ClearUI();

            _spacerLeft = MVRS.CreateSpacer(false);
            _spacerLeft.height = spacerHeight;
            _spacerRight = MVRS.CreateSpacer(true);
            _spacerRight.height = spacerHeight;

            GridLayoutGroup glg = _spacerLeft.gameObject.AddComponent<GridLayoutGroup>();
            glg.cellSize = new Vector2((1040 - ((buttonCount - 1) * buttonPadding)) / buttonCount, spacerHeight);

            glg.constraint = GridLayoutGroup.Constraint.FixedRowCount;
            glg.constraintCount = 1;
            glg.padding = new RectOffset(buttonPadding, buttonPadding, buttonPadding, buttonPadding);
            glg.spacing = new Vector2(buttonPadding, 0);

        }
    }

    public class TabSelectionUI : BaseMVRS
    {
        private ButtonRowUIElement topTabSelectionRow;
        private ButtonRowUIElement categorisationTabSelectionRow;
        private ButtonRowUIElement settingsTabSelectionRow;

        private LoadPluginsTabUI loadPluginsTabUI;
        private FilterPluginsTabUI filterPluginsTabUI;
        private PluginCategoriesTabUI pluginCategoriesTabUI;
        private CreatorCategoriesTabUI creatorCategoriesTabUI;
        private PluginFilterSettingsTabUI pluginFilterSettingsTabUI;
        private AboutSettingsTabUI aboutSettingsTabUI;
        private GeneralSettingsTabUI generalSettingsTabUI;
        private ModifyTagsCategoriesTabUI modifyTagsSettingsTabUI;
        private FindPluginsTabUI findPluginsTabUI;

        private BaseTabUI currentTabUI;

        private int currentTopTabSelection = -1;
        private int currentCatTabSelection = -1;
        private int currentSettingsTabSelection = -1;

        public TabSelectionUI(MVRScript mvrs) : base(mvrs)
        {
            topTabSelectionRow = new ButtonRowUIElement(MVRS);
            categorisationTabSelectionRow = new ButtonRowUIElement(MVRS);
            settingsTabSelectionRow = new ButtonRowUIElement(MVRS);

            filterPluginsTabUI = new FilterPluginsTabUI(MVRS);
            loadPluginsTabUI = new LoadPluginsTabUI(MVRS, filterPluginsTabUI);
            pluginCategoriesTabUI = new PluginCategoriesTabUI(MVRS, filterPluginsTabUI);
            creatorCategoriesTabUI = new CreatorCategoriesTabUI(MVRS, filterPluginsTabUI);
            pluginFilterSettingsTabUI = new PluginFilterSettingsTabUI(MVRS);
            modifyTagsSettingsTabUI = new ModifyTagsCategoriesTabUI(MVRS);
            findPluginsTabUI = new FindPluginsTabUI(MVRS);
            aboutSettingsTabUI = new AboutSettingsTabUI(MVRS);
            generalSettingsTabUI = new GeneralSettingsTabUI(MVRS);

            currentTabUI = loadPluginsTabUI;
            RefreshUI();
        }

        public void RefreshUI()
        {
            topTabSelectionRow.RefreshUI(4);

            topTabSelectionRow.AddButtonToRow("Load Plugins", TopMenuTabSelection.loadPlugins, TopTabMenuSelected, TopMenuTabSelection.loadPlugins);
            topTabSelectionRow.AddButtonToRow("Find Plugins", TopMenuTabSelection.findPlugins, TopTabMenuSelected, TopMenuTabSelection.loadPlugins);
            topTabSelectionRow.AddButtonToRow("Categorisation", TopMenuTabSelection.categories, TopTabMenuSelected, TopMenuTabSelection.loadPlugins);
            topTabSelectionRow.AddButtonToRow("Settings", TopMenuTabSelection.settings, TopTabMenuSelected, TopMenuTabSelection.loadPlugins);

            TopTabMenuSelected(TopMenuTabSelection.loadPlugins);
        }

        public void AtomNameUpdated(string oldName, string newName)
        {
            findPluginsTabUI.AtomNameUpdate( oldName, newName);
            loadPluginsTabUI.AtomNameUpdate(oldName, newName);
        }

        public void AtomRemoved(Atom atom)
        {
            findPluginsTabUI.AtomRemoved(atom);
            loadPluginsTabUI.AtomRemoved(atom);
        }

        public void TopTabMenuSelected(int buttonRef)
        {
            topTabSelectionRow.SetButtonColorActive(buttonRef);
            if (buttonRef != currentTopTabSelection)
            {
                currentTabUI.ClearAllGroupsUI(LeftRight.both,true);
                filterPluginsTabUI.ClearAllGroupsUI(LeftRight.both,true);

                if (buttonRef != TopMenuTabSelection.categories)  CloseAllCatTabUIs();

                switch (buttonRef)
                {
                    case TopMenuTabSelection.loadPlugins:
                        settingsTabSelectionRow.ClearUI();
                        categorisationTabSelectionRow.ClearUI();
                        filterPluginsTabUI.RefreshUI(true,false,true);
                        loadPluginsTabUI.RefreshUI(true);
                        currentTabUI = loadPluginsTabUI;
                        break;
                    case TopMenuTabSelection.findPlugins:
                        settingsTabSelectionRow.ClearUI();
                        categorisationTabSelectionRow.ClearUI();
                        findPluginsTabUI.RefreshUI(true);
                        currentTabUI = findPluginsTabUI;
                        break;
                    case TopMenuTabSelection.categories:
                        settingsTabSelectionRow.ClearUI();
                        categorisationTabSelectionRow.RefreshUI(3);
                        int tempCatTabSelection = currentCatTabSelection;
                        if (tempCatTabSelection == -1) tempCatTabSelection = CatMenuTabSelection.pluginCat;
                        categorisationTabSelectionRow.AddButtonToRow("Plugins", CatMenuTabSelection.pluginCat, CatTabMenuSelected, tempCatTabSelection);
                        categorisationTabSelectionRow.AddButtonToRow("Creators", CatMenuTabSelection.creatorCat, CatTabMenuSelected, tempCatTabSelection);
                        //                       categorisationTabSelectionRow.AddButtonToRow("Edit Types", CatMenuTabSelection.pluginTypes, CatTabMenuSelected, tempCatTabSelection);
                        categorisationTabSelectionRow.AddButtonToRow("Modify Plugin Tags", CatMenuTabSelection.pluginTags, CatTabMenuSelected, tempCatTabSelection);
                        
                        currentCatTabSelection = -1;
                        CatTabMenuSelected(tempCatTabSelection);
                        break;
                    case TopMenuTabSelection.settings:
                        categorisationTabSelectionRow.ClearUI();
                        settingsTabSelectionRow.RefreshUI(3);
                        int tempSettingsTabSelection = currentSettingsTabSelection;
                        if (tempSettingsTabSelection == -1) tempSettingsTabSelection = CatMenuTabSelection.pluginCat;
                        settingsTabSelectionRow.AddButtonToRow("Plugin Filter", SettingsMenuTabSelection.pluginFilter, SettingsTabMenuSelected, tempSettingsTabSelection);
                        settingsTabSelectionRow.AddButtonToRow("General", SettingsMenuTabSelection.general, SettingsTabMenuSelected, tempSettingsTabSelection);
                        settingsTabSelectionRow.AddButtonToRow("About", SettingsMenuTabSelection.about, SettingsTabMenuSelected, tempSettingsTabSelection);

                        currentSettingsTabSelection = -1;
                        SettingsTabMenuSelected(tempSettingsTabSelection) ;
                        break;
                }
                currentTopTabSelection = buttonRef;
            }
        }
        private void CatTabMenuSelected(int buttonRef)
        {
            categorisationTabSelectionRow.SetButtonColorActive(buttonRef);
            if (buttonRef != currentCatTabSelection)
            {
                currentTabUI.ClearAllGroupsUI(LeftRight.both,true);
                filterPluginsTabUI.ClearAllGroupsUI(LeftRight.both, true);
                switch (buttonRef)
                {
                    case CatMenuTabSelection.pluginCat:
                        filterPluginsTabUI.RefreshUI(true,true,false);
                        pluginCategoriesTabUI.RefreshUI(true);
                        currentTabUI = pluginCategoriesTabUI;
                        break;
                    case CatMenuTabSelection.creatorCat:
                        filterPluginsTabUI.RefreshUI(true, true, false,true);
                        creatorCategoriesTabUI.RefreshUI(true);
                        currentTabUI = creatorCategoriesTabUI;
                        break;
                    case CatMenuTabSelection.pluginTags:
                        modifyTagsSettingsTabUI.RefreshUI(true);
                        currentTabUI = modifyTagsSettingsTabUI;
                        break;
                    case CatMenuTabSelection.pluginTypes:
                        break;
                }
                currentCatTabSelection = buttonRef;
            }
        }

        private void SettingsTabMenuSelected(int buttonRef)
        {
            settingsTabSelectionRow.SetButtonColorActive(buttonRef);
            if (buttonRef != currentSettingsTabSelection)
            {
                currentTabUI.ClearAllGroupsUI(LeftRight.both, true);
                filterPluginsTabUI.ClearAllGroupsUI(LeftRight.both, true);
                switch (buttonRef)
                {
                    case SettingsMenuTabSelection.general:
                        generalSettingsTabUI.RefreshUI(true);
                        currentTabUI = generalSettingsTabUI;
                        break;
                    case SettingsMenuTabSelection.pluginFilter:
                        pluginFilterSettingsTabUI.RefreshUI(true);
                        currentTabUI = pluginFilterSettingsTabUI;
                        break;
                    case SettingsMenuTabSelection.about:
                        aboutSettingsTabUI.RefreshUI(true);
                        currentTabUI = aboutSettingsTabUI;
                        break;

                }
                currentSettingsTabSelection = buttonRef;
            }
        }
        private void CloseAllCatTabUIs()
        {
            pluginCategoriesTabUI.ClearAllGroupsUI(LeftRight.both,true);
        }
    }

    

    public abstract class BaseTabUI : BaseMVRS
    {
        protected class JSSCRefreshMap
        {
            public JSONStorableStringChooser jssc;
            public UIPopup.OnOpenPopup callback;
        }

        protected class JSSCCDVRefreshMap
        {
            public JSONStorableStringChooserCustomDisplayVal jsscCDV;
            public UIPopup.OnOpenPopup callback;
        }
        protected class GroupLR
        {
            public int groupID;
            public int leftRight;
        }

        public class UIDynamicElementLists
        {
            public HashSet<UIDynamic> UIDSpacersList;
            public HashSet<UIDynamicButton> UIDButtonsList;
            public HashSet<UIDynamicSlider> UIDSlidersList;
            public HashSet<UIDynamicPopup> UIDPopupList;
            public HashSet<UIDynamicColorPicker> UIDColorPickerList;
            public HashSet<UIDynamicTextField> UIDTextFieldList;
            public HashSet<UIDynamicToggle> UIDToggleList;

            public UIDynamicElementLists()
            {
                UIDSpacersList = new HashSet<UIDynamic>();
                UIDButtonsList = new HashSet<UIDynamicButton>();
                UIDSlidersList = new HashSet<UIDynamicSlider>();
                UIDPopupList = new HashSet<UIDynamicPopup>();
                UIDColorPickerList = new HashSet<UIDynamicColorPicker>();
                UIDTextFieldList = new HashSet<UIDynamicTextField>();
                UIDToggleList = new HashSet<UIDynamicToggle>();
            }

            public void ClearUIDs(MVRScript mvrs)
            {
                foreach (UIDynamic spacer in UIDSpacersList) if (spacer != null) mvrs.RemoveSpacer(spacer);
                foreach (UIDynamicButton button in UIDButtonsList) if (button != null) mvrs.RemoveButton(button);
                foreach (UIDynamicSlider slider in UIDSlidersList) if (slider != null) mvrs.RemoveSlider(slider);
                foreach (UIDynamicPopup popup in UIDPopupList) if (popup != null) mvrs.RemovePopup(popup);
                foreach (UIDynamicColorPicker colorpicker in UIDColorPickerList) if (colorpicker != null) mvrs.RemoveColorPicker(colorpicker);
                foreach (UIDynamicTextField textfield in UIDTextFieldList) if (textfield != null) mvrs.RemoveTextField(textfield);
                foreach (UIDynamicToggle toggle in UIDToggleList) if (toggle!=null) mvrs.RemoveToggle(toggle);
                UIDSpacersList.Clear();
                UIDButtonsList.Clear();
                UIDSlidersList.Clear();
                UIDPopupList.Clear();
                UIDColorPickerList.Clear();
                UIDTextFieldList.Clear();
                UIDToggleList.Clear();
            }

            public float GetUIElementsHeight()
            {
                float height = 0f;
                foreach (UIDynamic spacer in UIDSpacersList) height += spacer.height;
                foreach (UIDynamicButton button in UIDButtonsList) height += button.height;
                foreach (UIDynamicSlider slider in UIDSlidersList) height += slider.height;
                foreach (UIDynamicPopup popup in UIDPopupList) height += popup.height;
                foreach (UIDynamicColorPicker colorpicker in UIDColorPickerList) height += colorpicker.height;
                foreach (UIDynamicTextField textfield in UIDTextFieldList) height += textfield.height;
                foreach (UIDynamicToggle toggle in UIDToggleList) height += toggle.height;

                return height;
            }
            public int GetUIElementsCount()
            {
                int count = 0;
                foreach (UIDynamic spacer in UIDSpacersList) count += 1;
                foreach (UIDynamicButton button in UIDButtonsList) count += 1;
                foreach (UIDynamicSlider slider in UIDSlidersList) count += 1;
                foreach (UIDynamicPopup popup in UIDPopupList) count += 1;
                foreach (UIDynamicColorPicker colorpicker in UIDColorPickerList) count += 1;
                foreach (UIDynamicTextField textfield in UIDTextFieldList) count += 1;
                foreach (UIDynamicToggle toggle in UIDToggleList) count += 1;

                return count;
            }
            public void CloseAllPopup(UIDynamicPopup exception)
            {
                foreach (UIDynamicPopup UIDpopup in UIDPopupList)
                {
                    if (UIDpopup != exception) UIDpopup.popup.visible = false;
                }
            }
        }

        private Dictionary<int, UIDynamicElementLists> leftUIElementGroups;
        private Dictionary<int, UIDynamicElementLists> rightUIElementGroups;

        private Dictionary<string, List<JSSCRefreshMap>> jsscRefreshChoicesPreUIRefreshByGroupDict;
        private Dictionary<string, List<JSSCCDVRefreshMap>> jsscCDVRefreshChoicesPreUIRefreshByGroupDict;

        private Dictionary<JSONStorableStringChooser, HashSet<GroupLR>> jsscByGroupRefreshUIonValChangeDict;
        private Dictionary<JSONStorableBool, HashSet<GroupLR>> jsbByGroupRefreshUIonValChangeDict;
        private Dictionary<JSONStorableString, HashSet<GroupLR>> jssByGroupRefreshUIonValChangeDict;
        private Dictionary<JSONStorableFloat, HashSet<GroupLR>> jsfByGroupRefreshUIonValChangeDict;

        private Dictionary<JSONStorableStringChooser, JSONStorableStringChooser.SetJSONStringCallback> jsscOriginalCallbackOnValChangeDict;
        private Dictionary<JSONStorableBool, JSONStorableBool.SetJSONBoolCallback> jsbOriginalCallbackOnValChangeDict;
        private Dictionary<JSONStorableString, JSONStorableString.SetJSONStringCallback> jssOriginalCallbackOnValChangeDict;
        private Dictionary<JSONStorableFloat, JSONStorableFloat.SetJSONFloatCallback> jsfOriginalCallbackOnValChangeDict;

        protected bool activeUI = false;

        protected int uiGroupCount=1;

        public BaseTabUI(MVRScript mvrs) : base(mvrs)
        {
            leftUIElementGroups = new Dictionary<int, UIDynamicElementLists>();
            rightUIElementGroups = new Dictionary<int, UIDynamicElementLists>();
            jsscRefreshChoicesPreUIRefreshByGroupDict = new Dictionary<string, List<JSSCRefreshMap> >();
            jsscCDVRefreshChoicesPreUIRefreshByGroupDict = new Dictionary<string, List<JSSCCDVRefreshMap>>();

            jsscByGroupRefreshUIonValChangeDict = new Dictionary<JSONStorableStringChooser, HashSet<GroupLR>>();
            jsbByGroupRefreshUIonValChangeDict = new Dictionary<JSONStorableBool, HashSet<GroupLR>>();
            jssByGroupRefreshUIonValChangeDict = new Dictionary<JSONStorableString, HashSet<GroupLR>>();
            jsfByGroupRefreshUIonValChangeDict = new Dictionary<JSONStorableFloat, HashSet<GroupLR>>();

            jsscOriginalCallbackOnValChangeDict = new Dictionary<JSONStorableStringChooser, JSONStorableStringChooser.SetJSONStringCallback>();
            jsbOriginalCallbackOnValChangeDict = new Dictionary<JSONStorableBool, JSONStorableBool.SetJSONBoolCallback>();
            jssOriginalCallbackOnValChangeDict = new Dictionary<JSONStorableString, JSONStorableString.SetJSONStringCallback>();
            jsfOriginalCallbackOnValChangeDict = new Dictionary<JSONStorableFloat, JSONStorableFloat.SetJSONFloatCallback>();
        }
        protected void RegisterJSSCRefreshChoicesPreUIRefresh(JSONStorableStringChooser jssc, UIPopup.OnOpenPopup callback, int leftRightUI, int uiGroupID)
        {
            string key = leftRightUI.ToString() + ":" + uiGroupID.ToString();
            if (!jsscRefreshChoicesPreUIRefreshByGroupDict.ContainsKey(key)) jsscRefreshChoicesPreUIRefreshByGroupDict.Add(key, new List<JSSCRefreshMap>());
            JSSCRefreshMap jsscRefreshMap = new JSSCRefreshMap();
            jsscRefreshMap.jssc = jssc;
            jsscRefreshMap.callback = callback;
            jsscRefreshChoicesPreUIRefreshByGroupDict[key].Add(jsscRefreshMap);
        }
        protected void RegisterJSSCCDVRefreshChoicesPreUIRefresh(JSONStorableStringChooserCustomDisplayVal jsscCDV, UIPopup.OnOpenPopup callback, int leftRightUI, int uiGroupID)
        {
            string key = leftRightUI.ToString() + ":" + uiGroupID.ToString();
            if (!jsscCDVRefreshChoicesPreUIRefreshByGroupDict.ContainsKey(key)) jsscCDVRefreshChoicesPreUIRefreshByGroupDict.Add(key, new List<JSSCCDVRefreshMap>());
            JSSCCDVRefreshMap jsscCDVRefreshMap = new JSSCCDVRefreshMap();
            jsscCDVRefreshMap.jsscCDV = jsscCDV;
            jsscCDVRefreshMap.callback = callback;
            jsscCDVRefreshChoicesPreUIRefreshByGroupDict[key].Add(jsscCDVRefreshMap);
        }

        protected void RegisterJSSCRefreshUIonValChange(JSONStorableStringChooser jssc, int leftRight, int uiRefreshGroup)
        {
            GroupLR groupLR = new GroupLR();
            groupLR.groupID = uiRefreshGroup;
            groupLR.leftRight = leftRight;

            if (jssc.setJSONCallbackFunction != JSStringChooserCallback)
            {               
                if (jssc.setJSONCallbackFunction != null) jsscOriginalCallbackOnValChangeDict.Add(jssc, jssc.setJSONCallbackFunction);
                jssc.setJSONCallbackFunction = JSStringChooserCallback;
            }
      
            if (!jsscByGroupRefreshUIonValChangeDict.ContainsKey(jssc)) jsscByGroupRefreshUIonValChangeDict.Add(jssc, new HashSet<GroupLR>());
            jsscByGroupRefreshUIonValChangeDict[jssc].Add(groupLR);
        }

        protected void RegisterJSBRefreshUIonValChange(JSONStorableBool jsb, int leftRight, int uiRefreshGroup)
        {
            GroupLR groupLR = new GroupLR();
            groupLR.groupID = uiRefreshGroup;
            groupLR.leftRight = leftRight;

            if (jsb.setJSONCallbackFunction != JSBoolCallback)
            {
                if (jsb.setJSONCallbackFunction != null) jsbOriginalCallbackOnValChangeDict.Add(jsb, jsb.setJSONCallbackFunction);
                jsb.setJSONCallbackFunction = JSBoolCallback;
            }

            if (!jsbByGroupRefreshUIonValChangeDict.ContainsKey(jsb)) jsbByGroupRefreshUIonValChangeDict.Add(jsb, new HashSet<GroupLR>());
            jsbByGroupRefreshUIonValChangeDict[jsb].Add(groupLR);
        }

        protected void RegisterJSSRefreshUIonValChange(JSONStorableString jss, int leftRight, int uiRefreshGroup)
        {
            GroupLR groupLR = new GroupLR();
            groupLR.groupID = uiRefreshGroup;
            groupLR.leftRight = leftRight;

            if (jss.setJSONCallbackFunction != JSStringCallback)
            {
                if (jss.setJSONCallbackFunction != null) jssOriginalCallbackOnValChangeDict.Add(jss, jss.setJSONCallbackFunction);
                jss.setJSONCallbackFunction = JSStringCallback;
            }

            if (!jssByGroupRefreshUIonValChangeDict.ContainsKey(jss)) jssByGroupRefreshUIonValChangeDict.Add(jss, new HashSet<GroupLR>());
            jssByGroupRefreshUIonValChangeDict[jss].Add(groupLR);
        }

        protected void RegisterJSFRefreshUIonValChange(JSONStorableFloat jsf, int leftRight, int uiRefreshGroup)
        {
            GroupLR groupLR = new GroupLR();
            groupLR.groupID = uiRefreshGroup;
            groupLR.leftRight = leftRight;

            if (jsf.setJSONCallbackFunction != JSFloatCallback)
            {
                if (jsf.setJSONCallbackFunction != null) jsfOriginalCallbackOnValChangeDict.Add(jsf, jsf.setJSONCallbackFunction);
                jsf.setJSONCallbackFunction = JSFloatCallback;
            }

            if (!jsfByGroupRefreshUIonValChangeDict.ContainsKey(jsf)) jsfByGroupRefreshUIonValChangeDict.Add(jsf, new HashSet<GroupLR>());
            jsfByGroupRefreshUIonValChangeDict[jsf].Add(groupLR);
        }

        protected abstract void CreateGroupUI(int leftRight, int groupID);

        protected void RefreshGroupUI(int leftRight, int startGroupID)
        {
            if (leftRight == LeftRight.both)
            {
                RefreshGroupUI(LeftRight.left, startGroupID);
                RefreshGroupUI(LeftRight.right, startGroupID);
                return;
            }
            CloseAllPopup(leftRight, null);
            for (int groupID = startGroupID; groupID <= uiGroupCount; groupID++)
            {
                ClearGroupUI(leftRight, groupID);
            }
            for (int groupID = startGroupID; groupID <= uiGroupCount; groupID++)
            {
                string key = leftRight.ToString() + ":" + groupID.ToString();

                if (jsscRefreshChoicesPreUIRefreshByGroupDict.ContainsKey(key))
                {
                    foreach (JSSCRefreshMap jsscRefreshMap in jsscRefreshChoicesPreUIRefreshByGroupDict[key])
                    {
                        string originalDisplayVal="";
                        if (jsscRefreshMap.jssc.displayChoices != null && jsscRefreshMap.jssc.choices.Contains(jsscRefreshMap.jssc.val)) originalDisplayVal = jsscRefreshMap.jssc.displayChoices[jsscRefreshMap.jssc.choices.IndexOf(jsscRefreshMap.jssc.val)];
                        jsscRefreshMap.callback.Invoke();

                        if (jsscRefreshMap.jssc.displayChoices != null)
                        {
                            if (jsscRefreshMap.jssc.displayChoices.Count > 0 && !jsscRefreshMap.jssc.displayChoices.Contains(originalDisplayVal))
                            {
                                jsscRefreshMap.jssc.valNoCallback = jsscRefreshMap.jssc.choices[0];
                                if (jsscOriginalCallbackOnValChangeDict.ContainsKey(jsscRefreshMap.jssc)) jsscOriginalCallbackOnValChangeDict[jsscRefreshMap.jssc].Invoke(jsscRefreshMap.jssc);
                                if (jsscRefreshMap.jssc.setCallbackFunction != null) jsscRefreshMap.jssc.setCallbackFunction.Invoke(jsscRefreshMap.jssc.val);
                            }
                        }
                        else if (jsscRefreshMap.jssc.choices != null && !jsscRefreshMap.jssc.choices.Contains(jsscRefreshMap.jssc.val) && jsscRefreshMap.jssc.choices.Count > 0)
                        {
                            jsscRefreshMap.jssc.valNoCallback = jsscRefreshMap.jssc.choices[0];
                            if (jsscOriginalCallbackOnValChangeDict.ContainsKey(jsscRefreshMap.jssc)) jsscOriginalCallbackOnValChangeDict[jsscRefreshMap.jssc].Invoke(jsscRefreshMap.jssc);
                            if (jsscRefreshMap.jssc.setCallbackFunction != null) jsscRefreshMap.jssc.setCallbackFunction.Invoke(jsscRefreshMap.jssc.val);
                        }
                    }
                }
                if (jsscCDVRefreshChoicesPreUIRefreshByGroupDict.ContainsKey(key))
                {
                    foreach (JSSCCDVRefreshMap jsscCDVRefreshMap in jsscCDVRefreshChoicesPreUIRefreshByGroupDict[key])
                    {
                        string originalDisplayVal = "";
                        if (jsscCDVRefreshMap.jsscCDV.displayChoices != null && jsscCDVRefreshMap.jsscCDV.choices.Contains(jsscCDVRefreshMap.jsscCDV.val)) originalDisplayVal = jsscCDVRefreshMap.jsscCDV.displayChoices[jsscCDVRefreshMap.jsscCDV.choices.IndexOf(jsscCDVRefreshMap.jsscCDV.val)];
                        jsscCDVRefreshMap.callback.Invoke();

                        if (jsscCDVRefreshMap.jsscCDV.displayChoices != null && jsscCDVRefreshMap.jsscCDV.uniqueDisplayChoices)
                        {
                            if (!jsscCDVRefreshMap.jsscCDV.displayChoices.Contains(originalDisplayVal) && jsscCDVRefreshMap.jsscCDV.displayChoices.Count > 0)
                            {
                                jsscCDVRefreshMap.jsscCDV.valNoCallback = jsscCDVRefreshMap.jsscCDV.choices[0];
                                if (jsscOriginalCallbackOnValChangeDict.ContainsKey(jsscCDVRefreshMap.jsscCDV)) jsscOriginalCallbackOnValChangeDict[jsscCDVRefreshMap.jsscCDV].Invoke(jsscCDVRefreshMap.jsscCDV);
                                if (jsscCDVRefreshMap.jsscCDV.setCallbackFunction != null) jsscCDVRefreshMap.jsscCDV.setCallbackFunction.Invoke(jsscCDVRefreshMap.jsscCDV.val);
                            }
                        }
                        else if (jsscCDVRefreshMap.jsscCDV.choices != null && !jsscCDVRefreshMap.jsscCDV.choices.Contains(jsscCDVRefreshMap.jsscCDV.val) && jsscCDVRefreshMap.jsscCDV.choices.Count > 0)
                        {
                            jsscCDVRefreshMap.jsscCDV.valNoCallback = jsscCDVRefreshMap.jsscCDV.choices[0];
                            if (jsscOriginalCallbackOnValChangeDict.ContainsKey(jsscCDVRefreshMap.jsscCDV)) jsscOriginalCallbackOnValChangeDict[jsscCDVRefreshMap.jsscCDV].Invoke(jsscCDVRefreshMap.jsscCDV);
                            if (jsscCDVRefreshMap.jsscCDV.setCallbackFunction!=null) jsscCDVRefreshMap.jsscCDV.setCallbackFunction.Invoke(jsscCDVRefreshMap.jsscCDV.val);
                        }
                    }
                }
            }
            for (int groupID = startGroupID; groupID <= uiGroupCount; groupID++)
            {
                if (activeUI) CreateGroupUI(leftRight, groupID);
            }
        }

        private void JSStringChooserCallback(JSONStorableStringChooser jssc)
        {
            if (jsscOriginalCallbackOnValChangeDict.ContainsKey(jssc))
            {
                jsscOriginalCallbackOnValChangeDict[jssc].Invoke(jssc);
            }

            foreach (GroupLR groupLR in jsscByGroupRefreshUIonValChangeDict[jssc])
            {
                if (groupLR.groupID > 0)
                {
                    RefreshGroupUI(groupLR.leftRight, groupLR.groupID);
                }
            }


        }

        private void JSBoolCallback(JSONStorableBool jsb)
        {
            if (jsbOriginalCallbackOnValChangeDict.ContainsKey(jsb)) jsbOriginalCallbackOnValChangeDict[jsb].Invoke(jsb);
            foreach (GroupLR groupLR in jsbByGroupRefreshUIonValChangeDict[jsb])
            {
                if (groupLR.groupID > 0)
                {
                    RefreshGroupUI(groupLR.leftRight, groupLR.groupID);
                }
            }

        }

        private void JSStringCallback(JSONStorableString jss)
        {
            if (jssOriginalCallbackOnValChangeDict.ContainsKey(jss)) jssOriginalCallbackOnValChangeDict[jss].Invoke(jss);
            foreach (GroupLR groupLR in jssByGroupRefreshUIonValChangeDict[jss])
            {
                if (groupLR.groupID > 0)
                {
                    RefreshGroupUI(groupLR.leftRight, groupLR.groupID);
                }
            }

        }

        private void JSFloatCallback(JSONStorableFloat jsf)
        {
            if (jsfOriginalCallbackOnValChangeDict.ContainsKey(jsf)) jsfOriginalCallbackOnValChangeDict[jsf].Invoke(jsf);
            foreach (GroupLR groupLR in jsfByGroupRefreshUIonValChangeDict[jsf])
            {
                if (groupLR.groupID > 0)
                {
                    
                    RefreshGroupUI(groupLR.leftRight, groupLR.groupID);
                }
            }
        }

        private UIDynamicElementLists GetUIElementsFromGroup(int groupID, Dictionary<int, UIDynamicElementLists> uiElementsGroups)
        {
            UIDynamicElementLists uiElementGroup;

            if (uiElementsGroups.ContainsKey(groupID)) uiElementGroup = uiElementsGroups[groupID];
            else
            {
                uiElementGroup = new UIDynamicElementLists();
                uiElementsGroups.Add(groupID,uiElementGroup);
            }
            return (uiElementGroup);
        }

        protected UIDynamic CreateSpacer(float height = 10f,bool rightSide = false, int groupID=1)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamic spacer = MVRS.CreateSpacer(rightSide);
            if (rightSide) rightUIElements.UIDSpacersList.Add(spacer);
            else leftUIElements.UIDSpacersList.Add(spacer);
            spacer.height = height;
            return (spacer);
        }

        protected UIDynamicButton CreateButton(string label, bool rightSide = false, UnityEngine.Events.UnityAction callback =null, int groupID = 1)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamicButton button = MVRS.CreateButton(label, rightSide);
            if (rightSide) rightUIElements.UIDButtonsList.Add(button);
            else leftUIElements.UIDButtonsList.Add(button);

            if (callback!=null) button.button.onClick.AddListener(callback);

            return (button);
        }

        protected UIDynamicSlider CreateSlider(JSONStorableFloat jsonFloat, bool rightSide = false, int groupID = 1)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamicSlider slider = MVRS.CreateSlider(jsonFloat, rightSide);
            if (rightSide) rightUIElements.UIDSlidersList.Add(slider);
            else leftUIElements.UIDSlidersList.Add(slider);
            return (slider);
        }
        protected UIDynamicColorPicker CreateColorPicker(JSONStorableColor jsc, bool rightSide = false, int groupID = 1)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamicColorPicker colorPicker = MVRS.CreateColorPicker(jsc, rightSide);
            if (rightSide) rightUIElements.UIDColorPickerList.Add(colorPicker);
            else leftUIElements.UIDColorPickerList.Add(colorPicker);
            return (colorPicker);
        }
        protected UIDynamicToggle CreateToggle(JSONStorableBool jBool, bool rightSide = false, float height = 50f, int groupID = 1)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);
            UIDynamicToggle toggle = MVRS.CreateToggle(jBool, rightSide);
            toggle.height=height;
            if (rightSide) rightUIElements.UIDToggleList.Add(toggle);
            else leftUIElements.UIDToggleList.Add(toggle);
            return (toggle);
        }
        protected UIDynamicTextField CreateTextField(JSONStorableString jss, bool rightSide = false, bool input = false, InputField.LineType inputFieldType = InputField.LineType.MultiLineNewline, int groupID = 1)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamicTextField textfield = MVRS.CreateTextField(jss, rightSide);
            if (input)
            {
                textfield.height = 1f;
                textfield.backgroundColor = Color.white;
                var inputComponent = textfield.gameObject.AddComponent<InputField>();
                inputComponent.textComponent = textfield.UItext;
                inputComponent.lineType = inputFieldType;
                inputComponent.textComponent.resizeTextMaxSize = 30;
                inputComponent.textComponent.resizeTextForBestFit = true;
                jss.inputField = inputComponent;
            }

            if (rightSide) rightUIElements.UIDTextFieldList.Add(textfield);
            else leftUIElements.UIDTextFieldList.Add(textfield);
            return (textfield);
        }

        protected UIDynamicPopup CreatePopup(JSONStorableStringChooser stringChooserJSON, bool rightUI = false, float panelHeight = 400f, UIPopup.OnOpenPopup callbackMethod = null, string label = null, bool scrollable = true, bool closeOtherPopups=true, bool setValFromChoices = true, int groupID = 1)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamicPopup dynamicPopup;
            if (scrollable) dynamicPopup = MVRS.CreateScrollablePopup(stringChooserJSON, rightUI);
            else dynamicPopup = MVRS.CreatePopup(stringChooserJSON, rightUI);

            dynamicPopup.popup.showSlider = false;
            dynamicPopup.popupPanelHeight = panelHeight;

            int leftRightUI = LeftRight.left;
            if (rightUI) leftRightUI = LeftRight.right;

            if (callbackMethod != null)dynamicPopup.popup.onOpenPopupHandlers += callbackMethod;

            if (closeOtherPopups) dynamicPopup.popup.onOpenPopupHandlers += () =>
            {
                CloseAllPopup(leftRightUI, dynamicPopup);
            };

            if (setValFromChoices)
            {
                if (callbackMethod != null) callbackMethod.Invoke();
                if (stringChooserJSON.displayChoices != null) {
                    if (!stringChooserJSON.displayChoices.Contains(stringChooserJSON.val) && stringChooserJSON.displayChoices.Count>0)
                    {
                        stringChooserJSON.valNoCallback = stringChooserJSON.choices[0];
                    }
                 }
                else if (stringChooserJSON.choices != null && !stringChooserJSON.choices.Contains(stringChooserJSON.val) && stringChooserJSON.choices.Count > 0)
                {
                    stringChooserJSON.valNoCallback = stringChooserJSON.choices[0];
                }
            }

            if (label != null) { dynamicPopup.label = label; }

            if (rightUI) rightUIElements.UIDPopupList.Add(dynamicPopup);
            else leftUIElements.UIDPopupList.Add(dynamicPopup);

            return (dynamicPopup);
        }

        protected UIDynamicPopup CreateFilterablePopup(JSONStorableStringChooser stringChooserJSON, bool rightUI = false, float panelHeight = 400f, UIPopup.OnOpenPopup callbackMethod = null, string label = null, bool closeOtherPopups = true, int groupID = 1)
        {
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            UIDynamicPopup dynamicPopup = MVRS.CreateFilterablePopup(stringChooserJSON, rightUI);
            dynamicPopup.popupPanelHeight = panelHeight;

            int leftRightUI = LeftRight.left;
            if (rightUI) leftRightUI = LeftRight.right;

            if (callbackMethod != null) { dynamicPopup.popup.onOpenPopupHandlers += callbackMethod; }
            if (closeOtherPopups) dynamicPopup.popup.onOpenPopupHandlers += () =>
            {
                CloseAllPopup(leftRightUI, dynamicPopup);
            };

            if (label != null) { dynamicPopup.label = label; }

            if (rightUI) rightUIElements.UIDPopupList.Add(dynamicPopup);
            else leftUIElements.UIDPopupList.Add(dynamicPopup);

            return (dynamicPopup);
        }

        public void ClearAllGroupsUI(int leftRight, bool deactivateUI=false)
        {
            CloseAllPopup(leftRight, null);
            if (leftRight == LeftRight.right || leftRight == LeftRight.both)
            {
                foreach (KeyValuePair<int, UIDynamicElementLists> kvp in rightUIElementGroups) ClearGroupUI(leftRight,kvp.Key);
            }
            if (leftRight == LeftRight.left || leftRight == LeftRight.both)
            {
                foreach (KeyValuePair<int, UIDynamicElementLists> kvp in leftUIElementGroups) ClearGroupUI(leftRight, kvp.Key);
            }
            if (deactivateUI) activeUI = false;
        }
        public void ClearGroupUI(int leftRight, int groupID)
        {
            
            UIDynamicElementLists leftUIElements = GetUIElementsFromGroup(groupID, leftUIElementGroups);
            UIDynamicElementLists rightUIElements = GetUIElementsFromGroup(groupID, rightUIElementGroups);

            if (leftRight == LeftRight.right || leftRight == LeftRight.both) rightUIElements.ClearUIDs(MVRS);
            if (leftRight == LeftRight.left || leftRight == LeftRight.both) leftUIElements.ClearUIDs(MVRS);
        }
        public void CloseAllPopup(int leftRight, UIDynamicPopup exception)
        {
            if (leftRight == LeftRight.right || leftRight == LeftRight.both)
            {
                foreach (var rightUIElements in rightUIElementGroups.Values.ToList()) rightUIElements.CloseAllPopup(exception);
            }
            if (leftRight == LeftRight.left || leftRight == LeftRight.both)
            {
                foreach (var leftUIElements in leftUIElementGroups.Values.ToList()) leftUIElements.CloseAllPopup(exception);
            }
        }

        protected float GetTotalUIElementHeights(int leftRight)
        {
            Dictionary<int, UIDynamicElementLists> uiElementsGroup  ;
            
            if (leftRight == LeftRight.left) uiElementsGroup = leftUIElementGroups;
            else uiElementsGroup = rightUIElementGroups;

            float height = 0f;
            int count = 0;

            foreach(UIDynamicElementLists uidElementsList in uiElementsGroup.Values.ToList())
            {
                height += uidElementsList.GetUIElementsHeight();
                count += uidElementsList.GetUIElementsCount();
            }
            height += (count) * 15f;

            return (height);
        }

    }


    public class LoadPluginsTabUI : BaseTabUI
    {
        private JSONStorableString pluginDetailsJSON;
        private JSONStorableStringChooser targetAtomTypeJSON;
        private JSONStorableStringChooser targetAtomJSON;
        private JSONStorableBool openPluginUIJSON;

        private JSONStorableString pluginFileReferenceJSON;
        private JSONStorableString pmeNameJSON;

        private FilterPluginsTabUI pluginFiltersUI;

        public LoadPluginsTabUI(MVRScript mvrs, FilterPluginsTabUI _pluginFiltersUI) : base(mvrs)
        {
            pluginDetailsJSON = new JSONStorableString("pluginDetails", "");
            pluginFileReferenceJSON = new JSONStorableString("pluginFileRef", "", PluginFileUpdated);
            pmeNameJSON = new JSONStorableString("pmeName", "", PMENameUpdated);

            targetAtomTypeJSON = new JSONStorableStringChooser("targetAtomType", null, "1", "Target Atom Type");
            RegisterJSSCRefreshChoicesPreUIRefresh(targetAtomTypeJSON, AtomTypeChoicesInSceneRefresh, LeftRight.right, 1);
            RegisterJSSCRefreshUIonValChange(targetAtomTypeJSON, LeftRight.right, 2);

            targetAtomJSON = new JSONStorableStringChooser("targetAtom", null, "", "Target Atom");
            RegisterJSSCRefreshChoicesPreUIRefresh(targetAtomJSON, AtomChoicesInSceneRefresh, LeftRight.right, 2);

            openPluginUIJSON = new JSONStorableBool("Open Custom Plugin UI on Load", false);         

            pluginFiltersUI = _pluginFiltersUI;
            pluginFiltersUI.RegisterPluginFileNotification(pluginFileReferenceJSON);
            pluginFiltersUI.RegisterPMENotification(pmeNameJSON);

            uiGroupCount = 2;                       
        }
        public void RefreshUI(bool activateUI = false)
        {
            if (activateUI) activeUI = true;
            if (activeUI)
            {
                RefreshGroupUI(LeftRight.right, 1);
            }
        }

        protected override void CreateGroupUI(int leftRight, int groupID)
        {
            switch (groupID)
            {
                case 1:
                    CreateTextField(pluginDetailsJSON, true, false, InputField.LineType.MultiLineNewline, 1);
                    CreatePopup(targetAtomTypeJSON, true, 400, AtomTypeChoicesInSceneRefresh, null, true, true, false, 1);
                    break;
                case 2:
                    int atomType = int.Parse(targetAtomTypeJSON.val);
                    if (atomType != PluginAtomTypes.SessionPlugin && atomType != PluginAtomTypes.ScenePlugin)
                    {                       
                        CreateFilterablePopup(targetAtomJSON, true, 400, AtomChoicesInSceneRefresh, null, true, 2);
                    }                 

                    var button = CreateButton("Load Plugin to Target Atom", true, LoadPluginToTargetAtom, 2);
                    CreateToggle(openPluginUIJSON, true, 50, 2);
                    button.buttonColor = Color.green;

                    PluginAssist pa;
                    pa = (PluginAssist)MVRS;
                    if (!pa.settings.hideHelpText.val)
                    {
                        float height = Mathf.Max(0f, 840f - GetTotalUIElementHeights(LeftRight.right)+5f);
                        CreateSpacer(height, true, 2);
                        var textBox = CreateTextField(new JSONStorableString("", HelpText.loadPluginsHelpText), true, false, InputField.LineType.SingleLine, 2);
                        textBox.height = 270f;
                        textBox.backgroundColor = Color.grey;
                        textBox.textColor = Color.white;
                    }
                    break;
            }
        }   

        private void AtomTypeChoicesInSceneRefresh()
        {
            targetAtomTypeJSON.choices = PluginAtomTypes.GetMatchingTypes(SuperController.singleton.GetAtoms(),true,true,true);
            targetAtomTypeJSON.displayChoices = PluginAtomTypes.GetMatchingTypes(SuperController.singleton.GetAtoms(),true,false,true);
        }

        public void AtomNameUpdate(string oldName, string newName)
        {
            if (targetAtomJSON.val == oldName) targetAtomJSON.valNoCallback = newName;
        }
        public void AtomRemoved(Atom atom)
        {
            if (targetAtomJSON.val == atom.name)
            {
                RefreshGroupUI(LeftRight.right, 1);
            }
        }
        private void AtomChoicesInSceneRefresh()
        {
            targetAtomJSON.choices = PluginAtomTypes.GetMatchingAtomNames(SuperController.singleton.GetAtoms(), int.Parse(targetAtomTypeJSON.val));
        }      

        private void LoadPluginToTargetAtom()
        {
            string atomName = targetAtomJSON.val;
            string targetAtomType = PluginAtomTypes.Label(int.Parse(targetAtomTypeJSON.val));
            if (targetAtomType == "Scene Plugin" || targetAtomType == "Session Plugin") atomName = "CoreControl";
            Atom atom = SuperController.singleton.GetAtomByUid(atomName);
            if (targetAtomType == "Session Plugin") atom = MVRS.containingAtom;
            LoadPluginsAsPreset(atom);
        }

        private void PluginFileUpdated(string val)
        {
            pluginDetailsJSON.val = val;
        }

        private void PMENameUpdated(string val)
        {
            PluginAssist pa = (PluginAssist)MVRS;
            PluginManifestEntry pme = pa.pluginManifest.GetPluginManifestEntry(val);

            AtomTypeChoicesInSceneRefresh();
            if (pme.pluginType!=PluginAtomTypes.All && targetAtomTypeJSON.choices.Contains(pme.pluginType.ToString())) targetAtomTypeJSON.val = pme.pluginType.ToString();
        }
        private void LoadPluginsAsPreset(Atom atom)
        {
            MVRPluginManager manager;

            if (atom.type == "SessionPluginManager") manager = MVRS.manager;
            else manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;

            JSONClass pluginMgrJC = manager.GetJSON(true, true, true);
            JSONClass pluginsJC = pluginMgrJC["plugins"].AsObject;
            int nextFreeSlot = pluginsJC.Count;

            JSONClass pluginPresetJSON = new JSONClass();
            pluginPresetJSON["setUnlistedParamsToDefault"].AsBool = true;

            JSONArray pluginPresetStorablesJSON = new JSONArray();
            pluginPresetJSON["storables"] = pluginPresetStorablesJSON;

            JSONClass pluginMgrStorableJSON = new JSONClass();
            pluginPresetStorablesJSON.Add(pluginMgrStorableJSON);
            pluginMgrStorableJSON["id"] = "PluginManager";

            JSONClass pluginMgrPluginsListJSON = new JSONClass();
            pluginMgrStorableJSON["plugins"] = pluginMgrPluginsListJSON;

            string normalFilePath = SuperController.singleton.NormalizePath(pluginFileReferenceJSON.val);
            if (FileManagerSecure.FileExists(pluginFileReferenceJSON.val))
            {
                pluginMgrPluginsListJSON["plugin#0"] = normalFilePath;
            }

            PresetManager pm;
            if (atom.name == "CoreControl")
            {
                if (atom.type == "SessionPluginManager") pm = atom.GetComponentInChildren<PresetManager>();
                else
                {
                    JSONStorable js = atom.GetStorableByID("PluginManagerPresets");
                    pm = js.GetComponentInChildren<PresetManager>();
                }
            }
            else
            {
                if (atom.type != "Person") pm = atom.GetComponentInChildren<PresetManager>();
                else
                {
                    JSONStorable js = atom.GetStorableByID("PluginPresets");
                    pm = js.GetComponentInChildren<PresetManager>();
                }
            }

            atom.SetLastRestoredData(pluginPresetJSON, true, true);
            pm.LoadPresetFromJSON(pluginPresetJSON,true);

            if (openPluginUIJSON.val) Utils.SelectAndOpenUI(atom, nextFreeSlot);
        }

     }

    public class FindPluginsTabUI : BaseTabUI
    {
        private JSONStorableStringChooser pluginSelectorJSON;
        private JSONStorableFixedIndexStringChoooser atomSelectorJSON;
        private JSONStorableStringChooser pluginSlotSelectorJSON;
        private JSONStorableStringChooser pluginUISelectorJSON;
        private ScenePluginManifest scenePluginManifest;

        private Atom currentAtom;

        public FindPluginsTabUI(MVRScript mvrs) : base(mvrs)
        {
            PluginAssist pa = (PluginAssist)mvrs;
            scenePluginManifest = pa.scenePluginManifest;

            pluginSelectorJSON = new JSONStorableStringChooser("pluginSelector", null, "", "Current Scene Plugins");
            RegisterJSSCRefreshChoicesPreUIRefresh(pluginSelectorJSON, PluginChoicesInSceneRefresh, LeftRight.left, 1);
            RegisterJSSCRefreshUIonValChange(pluginSelectorJSON, LeftRight.left, 2);
            RegisterJSSCRefreshUIonValChange(pluginSelectorJSON, LeftRight.right, 1);

            atomSelectorJSON = new JSONStorableFixedIndexStringChoooser("atomSelector", null,null, "", "Containing Atom",FixedIndexAtomSelected, DynamicIndexAtomSelected);
            RegisterJSSCCDVRefreshChoicesPreUIRefresh(atomSelectorJSON, AtomChoicesInSceneRefresh, LeftRight.left, 2);
            RegisterJSSCRefreshUIonValChange(atomSelectorJSON, LeftRight.left, 3);
            RegisterJSSCRefreshUIonValChange(atomSelectorJSON, LeftRight.right, 1);

            pluginSlotSelectorJSON = new JSONStorableStringChooser("pluginSlotSelector", null, "", "Plugin Slot");
            RegisterJSSCRefreshChoicesPreUIRefresh(pluginSlotSelectorJSON, PluginSlotChoicesRefresh, LeftRight.left, 3);

            pluginUISelectorJSON = new JSONStorableStringChooser("pluginUISelector", null, "", "Select Plugin UI");
            RegisterJSSCRefreshChoicesPreUIRefresh(pluginUISelectorJSON, PluginUIChoicesInSceneRefresh, LeftRight.right, 1);

            uiGroupCount = 3;
        }
        public void RefreshUI(bool activateUI = false)
        {
            if (activateUI) activeUI = true;
            if (activeUI)
            {
                RefreshGroupUI(LeftRight.both, 1);
            }
        }

        protected override void CreateGroupUI(int leftRight, int groupID)
        {
            PluginAssist pa;
            pa = (PluginAssist)MVRS;

            if (leftRight == LeftRight.left || leftRight == LeftRight.both)
            {
                switch (groupID)
                {
                    case 1:
                        CreateFilterablePopup(pluginSelectorJSON, false, 400, PluginChoicesInSceneRefreshManifest, null, true, 1);
                        break;
                    case 2:
                        CreateFilterablePopup(atomSelectorJSON, false, 400, AtomChoicesInSceneRefreshManifest, null, true, 2);
                        break;
                    case 3:
                        CreatePopup(pluginSlotSelectorJSON, false, 400, PluginSlotChoicesRefreshManifest, null, true, true, true, 3);

                        if (!pa.settings.hideHelpText.val)
                        {
                            float height = Mathf.Max(0f, 840f - GetTotalUIElementHeights(LeftRight.left));
                            CreateSpacer(height, false, 3);
                            var textBox = CreateTextField(new JSONStorableString("", HelpText.findPluginsLeftHelpText), false, false, InputField.LineType.SingleLine, 3);
                            textBox.height = 270f;
                            textBox.backgroundColor = Color.grey;
                            textBox.textColor = Color.white;
                        }

                        break;

                }
            }
            if (leftRight == LeftRight.right || leftRight == LeftRight.both)
            {
                switch (groupID)
                {
                    case 1:
                        if (pluginUISelectorJSON.choices != null && pluginUISelectorJSON.choices.Count > 1)
                        {
                            CreatePopup(pluginUISelectorJSON, true, 400, PluginUIChoicesInSceneRefresh, null, true, true, true, 1);
                        }
                        var button = CreateButton("Open Plugin UI",true, OpenUI, 1);
                        button.buttonColor = Color.green;

                        if (!pa.settings.hideHelpText.val)
                        {
                            float height = Mathf.Max(0f, 840f - GetTotalUIElementHeights(LeftRight.right) + 5f);
                            CreateSpacer(height, true, 1);
                            var textBox = CreateTextField(new JSONStorableString("", HelpText.findPluginsRightHelpText), true, false, InputField.LineType.SingleLine, 1);
                            textBox.height = 270f;
                            textBox.backgroundColor = Color.grey;
                            textBox.textColor = Color.white;
                        }
                        break;
                }
            }
        }

        private void PluginChoicesInSceneRefreshManifest()
        {
            scenePluginManifest.RefreshManifest();
            PluginChoicesInSceneRefresh();
        }
        private void AtomChoicesInSceneRefreshManifest()
        {
            scenePluginManifest.RefreshManifest();
            AtomChoicesInSceneRefresh();
        }
        private void PluginSlotChoicesRefreshManifest()
        {
            scenePluginManifest.RefreshManifest();
            PluginSlotChoicesRefresh();
        }
        public void AtomRemoved(Atom atom)
        {
            if (!atomSelectorJSON.isValFixedIndex && atomSelectorJSON.displayVal == atom.name)
            {
                RefreshGroupUI(LeftRight.left, 1);
            }
        }

        public void AtomNameUpdate(string oldName, string newName)
        {
            if(!atomSelectorJSON.isValFixedIndex && atomSelectorJSON.displayVal== oldName)
            {
                RefreshGroupUI(LeftRight.left, 2);
            }
        }

        private Atom GetCurrentAtom()
        {
            Atom currentAtom;

            if (atomSelectorJSON.isValFixedIndex)
            {
                if (atomSelectorJSON.displayVal == "Session Plugins") currentAtom = MVRS.containingAtom;
                else currentAtom = SuperController.singleton.GetAtomByUid("CoreControl");
            }
            else
            {
                currentAtom = SuperController.singleton.GetAtomByUid(atomSelectorJSON.displayVal);
            }
            return currentAtom;
        }

        private void PluginSlotChoicesRefresh()
        {
            ScenePluginManifestEntry sPME = scenePluginManifest.GetScenePluginManifestEntry(pluginSelectorJSON.val);
            Atom currentAtom = GetCurrentAtom();

            List<string> pluginSlotChoices = sPME.GetPluginSlotsFromAtom(currentAtom);
            pluginSlotSelectorJSON.choices = pluginSlotChoices;

        }
        private void PluginUIChoicesInSceneRefresh()
        {          
            List<string> choices = new List<string>();
            Atom currentAtom = GetCurrentAtom();

            MVRPluginManager manager;
            if (currentAtom.type == "SessionPluginManager") manager = MVRS.manager;
            else manager = currentAtom.GetStorableByID("PluginManager") as MVRPluginManager;

            foreach (Transform transform in manager.gameObject.transform.Find("Plugins"))
            {
                MVRScript script = transform.gameObject.GetComponent<MVRScript>();
                if (script.name.StartsWith(pluginSlotSelectorJSON.val)) choices.Add(script.name.Substring(pluginSlotSelectorJSON.val.Length+1));
            }
            pluginUISelectorJSON.choices = choices;
        }

        private void PluginChoicesInSceneRefresh()
        {
            List<string> pluginNameChoices = scenePluginManifest.GetPluginsFromScene();
            List<string> pluginPMENameChoices = scenePluginManifest.GetPluginsFromScene(true);

            Utils.SortLists(ref pluginNameChoices, ref pluginPMENameChoices);
            pluginSelectorJSON.choices = pluginPMENameChoices;
            pluginSelectorJSON.displayChoices = pluginNameChoices;
        }


        private void AtomChoicesInSceneRefresh()
        {
            if (pluginSelectorJSON.val != "")
            {
                ScenePluginManifestEntry sPME = scenePluginManifest.GetScenePluginManifestEntry(pluginSelectorJSON.val);
                List<Atom> atoms = sPME.GetAtomList();
                List<string> fixedIndexChoices = new List<string>();
                List<string> dynamicIndexChoices = new List<string>();

                foreach (Atom atom in atoms)
                {
                    if (atom.name == "CoreControl")
                    {
                        if (atom.type == "SessionPluginManager") fixedIndexChoices.Add("Session Plugins");
                        else fixedIndexChoices.Add("Scene Plugins");
                    }
                    else dynamicIndexChoices.Add(atom.name);
                }
                atomSelectorJSON.SetIndexChoices(fixedIndexChoices, dynamicIndexChoices);
            }

        }

        private void FixedIndexAtomSelected(JSONStorableStringChooser atomJSON)
        {
        }
        private void DynamicIndexAtomSelected(JSONStorableStringChooser atomJSON)
        {
        }

        private void OpenUI()
        {
            Utils.SelectAndOpenUI(GetCurrentAtom(),int.Parse(pluginSlotSelectorJSON.val.Substring(7)),pluginUISelectorJSON.val);
        }

     }
    public class PluginCategoriesTabUI : BaseTabUI
    {
        private JSONStorableString pluginDetailsJSON;
        private PluginManifestEntry pme;

        private FilterPluginsTabUI pluginFiltersUI;
        private JSONStorableString pluginFileReferenceJSON;
        private JSONStorableStringChooser pluginTypeSelectionJSON;
        private JSONStorableBool pluginPreferredJSON;
        private JSONStorableBool pluginFolderVariantConcealedJSON;
        private JSONStorableString pmeNameJSON;
        private JSONStorableTagChooser tagChooserJSON;

        private string folderVariant = "";        

        public PluginCategoriesTabUI(MVRScript mvrs, FilterPluginsTabUI _pluginFiltersUI) : base(mvrs)
        {
            pluginDetailsJSON = new JSONStorableString("pluginDetails", "");
            pmeNameJSON = new JSONStorableString("pmeName", "",PMENameUpdated);
            pluginFileReferenceJSON = new JSONStorableString("pluginFileReference", "", PluginFileUpdated);           

            pluginTypeSelectionJSON = new JSONStorableStringChooser("pluginType", null, "1", "Set Plugin Type", PluginTypeSelected);
            pluginPreferredJSON = new JSONStorableBool("Preferred Plugin", false, PluginPreferSelected);
            RegisterJSBRefreshUIonValChange(pluginPreferredJSON,LeftRight.right, 2);
            pluginFolderVariantConcealedJSON =new JSONStorableBool("Conceal Plugin", false, PluginConcealSelected);
            RegisterJSBRefreshUIonValChange(pluginFolderVariantConcealedJSON, LeftRight.right, 2);           

            tagChooserJSON = new JSONStorableTagChooser("pluginTagChooser",null,"","Set Plugin Tags",PluginTagSelected,pmeNameJSON,mvrs);

            pluginFiltersUI = _pluginFiltersUI;
            pluginFiltersUI.RegisterPluginFileNotification(pluginFileReferenceJSON);
            pluginFiltersUI.RegisterPMENotification(pmeNameJSON);

            uiGroupCount = 2;
        }
        public void RefreshUI(bool activateUI=false)
        {
            if (activateUI) activeUI = true;
            if (activeUI)
            {
                RefreshGroupUI(LeftRight.right,1);
            }
        }

        protected override void CreateGroupUI(int leftRight,int groupID)
        {
            PluginAssist pa;
            pa = (PluginAssist)MVRS;

            switch (groupID)
            {
                case 1:
                    CreateTextField(pluginDetailsJSON, true, false, InputField.LineType.MultiLineNewline, 1);
                    UIDynamicPopup popup = CreatePopup(pluginTypeSelectionJSON, true, 600, PluginTypeChoicesRefresh, null, true, true, false, 1);
                    popup.labelTextColor = Color.green;
                    var toggle1 = CreateToggle(pluginPreferredJSON, true);
                    if (pme.GetPathVariantsInSameVARCount() > 1) toggle1.label = "Set Preferred Plugin Variant";
                    else toggle1.label = "Set Preferred Plugin";
                    toggle1.backgroundColor = Color.green;

                    var toggle2 = CreateToggle(pluginFolderVariantConcealedJSON, true);
                    if (pme.GetPathVariantsInSameVARCount() > 1) toggle2.label = "Set Conceal Plugin Variant";
                    else toggle2.label = "Set Conceal Plugin";
                    toggle2.backgroundColor = Color.green;

                    break;
                case 2:
                    if (pme.GetConcealedPathVariantCount() < (pme.GetPathVariantsInSameVARCount() - 1) && !pluginFolderVariantConcealedJSON.val)
                    {
                        var button = CreateButton("Set other Folder Variants as Concealed", true, SetOtherPathVariantsConcealed, 2);
                    }
                    if (PC.patreonContentEnabled)
                    {
                        popup = CreateFilterablePopup(tagChooserJSON, true, 600, TagChooserChoicesRefresh, null, true, 2);
                        popup.labelTextColor = Color.green;
                        tagChooserJSON.valNoCallback = pme.pluginTagSet.GetTagSetSummary();
                    }

                    if (!pa.settings.hideHelpText.val)
                    {
                        float height = Mathf.Max(0f, 840f - GetTotalUIElementHeights(LeftRight.right)+5f-65f);
                        CreateSpacer(height, true, 2);
                        var textBox = CreateTextField(new JSONStorableString("", HelpText.catPluginsHelpText), true, false, InputField.LineType.SingleLine, 2);
                        textBox.height = 270f;
                        textBox.backgroundColor = Color.grey;
                        textBox.textColor = Color.white;
                    }
                    break;
            }
        }

        private void TagChooserChoicesRefresh()
        {
            PluginAssist pa = (PluginAssist)MVRS;
            List<string> choices = pa.tagsManifest.GetTagLabels();
            choices.Sort();
            tagChooserJSON.choices = choices;
            tagChooserJSON.displayChoices = choices;
        }
        private void PluginPreferSelected(bool val)
        {
            pme.SetVariantPreferred(folderVariant, pluginPreferredJSON.val);
        }
        private void PluginConcealSelected(bool val)
        {
            pme.SetVariantConcealed(folderVariant, pluginFolderVariantConcealedJSON.val);
        }
        private void PluginTypeSelected(string val)
        {
            pme.pluginType = int.Parse(pluginTypeSelectionJSON.val);
        }
        private void PluginTagSelected(string val)
        {
            PluginAssist pa = (PluginAssist)MVRS;
            PluginTag tag = pa.tagsManifest.GetTag(val);
            pme.pluginTagSet.ToggleTagValue(tag);
            tagChooserJSON.valNoCallback = pme.pluginTagSet.GetTagSetSummary();
        }

        private void SetOtherPathVariantsConcealed()
        {
            pme.SetAllVariantConcealedExcept(folderVariant);
            RefreshGroupUI(LeftRight.right, 2);
        }

        private void PluginTypeChoicesRefresh()
        {
            pluginTypeSelectionJSON.choices = PluginAtomTypes.GetAllTypes(true);
            pluginTypeSelectionJSON.displayChoices = PluginAtomTypes.GetAllTypes(false);
        }

        private void PluginFileUpdated(string val)
        {
            pluginDetailsJSON.val = val;
            folderVariant = Utils.GetFolderFromPath(val);
            if (pme != null)
            {
                pluginPreferredJSON.valNoCallback = pme.GetVariantPreferred(folderVariant);
                pluginFolderVariantConcealedJSON.valNoCallback = pme.GetVariantConcelealed(folderVariant);
            }
            RefreshGroupUI(LeftRight.right, 1);
        }
        private void PMENameUpdated(string val)
        {
            PluginAssist pa = (PluginAssist)MVRS;
            pme = pa.pluginManifest.GetPluginManifestEntry(val);
            PluginTypeChoicesRefresh();
            pluginTypeSelectionJSON.valNoCallback = pme.pluginType.ToString();
        }
        
    }

    public class CreatorCategoriesTabUI : BaseTabUI
    {

        private FilterPluginsTabUI pluginFiltersUI;

        private JSONStorableBool creatorPreferredJSON;
        private JSONStorableBool creatorConcealedJSON;

        private JSONStorableString creatorNameJSON;
        private JSONStorableBool creatorNameIsCollectionJSON;

        public CreatorCategoriesTabUI(MVRScript mvrs, FilterPluginsTabUI _pluginFiltersUI) : base(mvrs)
        {
            creatorPreferredJSON = new JSONStorableBool("Set Preferred Creator", false, CreatorPreferredSelected);
            creatorConcealedJSON = new JSONStorableBool("Conceal Creator & all their Plugins", false, CreatorConcealSelected);
            creatorNameJSON = new JSONStorableString("CreatorName", "",CreatorNameUpdated);
            creatorNameIsCollectionJSON = new JSONStorableBool("CreatorNameIsCollection", false);

            pluginFiltersUI = _pluginFiltersUI;
            pluginFiltersUI.RegisterCreatorNameNotification(creatorNameJSON, creatorNameIsCollectionJSON);

            uiGroupCount = 1;
        }
        protected override void CreateGroupUI(int leftRight,int groupID)
        {
            PluginAssist pa;
            pa = (PluginAssist)MVRS;

            switch (groupID)
            {
                case 1:
                    UIDynamicToggle toggle1= CreateToggle(creatorPreferredJSON, true);
                    toggle1.backgroundColor = Color.green;
                    UIDynamicToggle toggle2 = CreateToggle(creatorConcealedJSON, true);
                    toggle2.backgroundColor = Color.green;
                    if (!pa.settings.hideHelpText.val)
                    {
                        float height = Mathf.Max(0f, 840f - GetTotalUIElementHeights(LeftRight.right) + 5f -65f);
                        CreateSpacer(height, true, 1);
                        var textBox = CreateTextField(new JSONStorableString("", HelpText.catCreatorsHelpText), true, false, InputField.LineType.SingleLine, 1);
                        textBox.height = 270f;
                        textBox.backgroundColor = Color.grey;
                        textBox.textColor = Color.white;
                    }
                    break;
            }
        }
        public void RefreshUI(bool activateUI = false)
        {
            if (activateUI) activeUI = true;
            if (activeUI)
            {
                RefreshGroupUI(LeftRight.right, 1);
            }
        }

        private void CreatorPreferredSelected(bool val)
        {
            PluginAssist pa = (PluginAssist)MVRS;
            pa.pluginManifest.SetCreatorPreferred(creatorNameJSON.val,val);
        }
        private void CreatorConcealSelected(bool val)
        {
            PluginAssist pa = (PluginAssist)MVRS;
            pa.pluginManifest.SetCreatorConcealed(creatorNameJSON.val, val);
        }

        private void CreatorNameUpdated(string val)
        {
            if (!creatorNameIsCollectionJSON.val)
            {
                PluginAssist pa = (PluginAssist)MVRS;
                creatorConcealedJSON.valNoCallback = pa.pluginManifest.GetCreatorConcealed(val);
                creatorPreferredJSON.valNoCallback = pa.pluginManifest.GetCreatorPreferred(val);
//                RefreshGroupUI(LeftRight.right, 1);
            }

        }

    }

    public class FilterPluginsTabUI : BaseTabUI
    {
        private JSONStorableCreatorChooser creatorFilterJSON;
        private JSONStorableStringChooser pluginTypeFilterJSON;
        private JSONStorableFixedIndexStringChoooser tagsFilterJSON;
        private JSONStorablePluginChooser pluginsFilterJSON;
        private JSONStorableStringChooser packageVersionsFilterJSON;
        private JSONStorablePathVariantChooser pathVariantFilterJSON;

        private JSONStorableBool showConcealedCreatorsJSON;
        private JSONStorableBool showConcealedPluginsJSON;

        private string pluginFileReference="";

        private HashSet<JSONStorableString> pluginFileNotiticationList;
        private HashSet<JSONStorableString> pmeNotiticationList;
        private HashSet<JSONStorableString> creatorNameNotificationList;
        private HashSet<JSONStorableBool> creatorNameIsCollectionNotificationList;
        private HashSet<JSONStorableString> pluginTypeFilterNotiticationList;

        private bool showConcealedOption = false;
        private bool showVARVersions = false;
        private bool hideCreatorGroups = false;

        private bool allCreators
        {
            get
            {
                return (creatorFilterJSON.isValFixedIndex && (creatorFilterJSON.displayVal == "ALL" || creatorFilterJSON.displayVal == "VAR ONLY"));
            }
        }
        private bool localCreators
        {
            get { return creatorFilterJSON.isValFixedIndex && (creatorFilterJSON.displayVal == "LOCAL ONLY" || creatorFilterJSON.displayVal == "ALL"); }

         }

        public FilterPluginsTabUI(MVRScript mvrs) : base(mvrs)
        {
            List<string> fixedIndexCreators = new List<string>();
            fixedIndexCreators.Add("ALL");
            fixedIndexCreators.Add("VAR ONLY");
            fixedIndexCreators.Add("LOCAL ONLY");

            List<string> fixedIndexTags = new List<string>();
            fixedIndexTags.Add("All");

            PluginAssist pa = (PluginAssist)MVRS;

            //UI Group 1
            creatorFilterJSON = new JSONStorableCreatorChooser("creator", fixedIndexCreators, null, "0", "Creator Filter", FixedIndexCreatorSelected, CreatorSelected,pa );
            RegisterJSSCCDVRefreshChoicesPreUIRefresh(creatorFilterJSON, CreatorChoicesRefresh, LeftRight.left, 1);
            RegisterJSSCRefreshUIonValChange(creatorFilterJSON, LeftRight.left, 2);

            showConcealedCreatorsJSON = new JSONStorableBool("Show Concealed Creators", false);
            RegisterJSBRefreshUIonValChange(showConcealedCreatorsJSON, LeftRight.left, 1);

            //UI Group 2
            pluginTypeFilterJSON = new JSONStorableStringChooser("atomType", null, "0", "Plugin Type Filter", PluginTypeSelected);
            RegisterJSSCRefreshChoicesPreUIRefresh(pluginTypeFilterJSON, PluginTypeFilterChoicesRefresh, LeftRight.left, 2);
            RegisterJSSCRefreshUIonValChange(pluginTypeFilterJSON, LeftRight.left, 3);

            //UI Group 3
            tagsFilterJSON = new JSONStorableFixedIndexStringChoooser("tags", fixedIndexTags, null, "All", "Plugin Tags Filter", FixedIndexTagSelected, TagSelected);
            if (PC.patreonContentEnabled)
            {
                RegisterJSSCCDVRefreshChoicesPreUIRefresh(tagsFilterJSON, PluginTagFilterChoicesRefresh, LeftRight.left, 3);
                RegisterJSSCRefreshUIonValChange(tagsFilterJSON, LeftRight.left, 4);
            }

            

            //UI Group 4
            pluginsFilterJSON = new JSONStorablePluginChooser("plugins", null, "", "Plugin Selector",PluginSelected,pa);
            RegisterJSSCCDVRefreshChoicesPreUIRefresh(pluginsFilterJSON, PluginChoicesRefresh, LeftRight.left, 4);
            RegisterJSSCRefreshUIonValChange(pluginsFilterJSON, LeftRight.left, 5);

            //UI Group 5
            packageVersionsFilterJSON = new JSONStorableStringChooser("packageVersions", null, "", "Package Version Selector");
            RegisterJSSCRefreshChoicesPreUIRefresh(packageVersionsFilterJSON, PackageVersionChoicesRefresh, LeftRight.left, 5);
            RegisterJSSCRefreshUIonValChange(packageVersionsFilterJSON, LeftRight.left, 6);

            // UI Group 6
            pathVariantFilterJSON = new JSONStorablePathVariantChooser("pathVariants", null, "", "Folder Variant Selector",PathVariantSelected,pa,pluginsFilterJSON);
            RegisterJSSCCDVRefreshChoicesPreUIRefresh(pathVariantFilterJSON, PathVariantChoicesRefresh, LeftRight.left, 6);
            showConcealedPluginsJSON = new JSONStorableBool("Show Concealed Plugins", false);
            RegisterJSBRefreshUIonValChange(showConcealedPluginsJSON, LeftRight.left, 4);

            uiGroupCount = 6;

            pluginFileNotiticationList = new HashSet<JSONStorableString>();
            pmeNotiticationList = new HashSet<JSONStorableString>();
            creatorNameNotificationList = new HashSet<JSONStorableString>();
            creatorNameIsCollectionNotificationList = new HashSet<JSONStorableBool>();
            pluginTypeFilterNotiticationList = new HashSet<JSONStorableString>();
        }

        private void PluginSelected (string pluginPMEName)
        {
            PluginAssist pa = (PluginAssist)MVRS;
            PluginManifestEntry pme = pa.pluginManifest.GetPluginManifestEntry(pluginPMEName);
            if (!pme.isLocal && pme.varVersionCount > 1) packageVersionsFilterJSON.valNoCallback = pme.highestVARNumber.ToString();
            
        }

        public void RegisterPluginTypeNotification(JSONStorableString pluginTypeJSON)
        {
            pluginTypeFilterNotiticationList.Add(pluginTypeJSON);
            pluginTypeJSON.val = pluginTypeFilterJSON.val;
        }

        public void RegisterPluginFileNotification(JSONStorableString pluginFileJSON)
        {
            pluginFileNotiticationList.Add(pluginFileJSON);
            pluginFileJSON.val = pluginFileReference;
        }

        public void DeregisterPluginFileNotification(JSONStorableString pluginFileJSON)
        {
            pluginFileNotiticationList.Remove(pluginFileJSON);
        }

        public void RegisterPMENotification(JSONStorableString pmeJSON)
        {
            pmeNotiticationList.Add(pmeJSON);
            pmeJSON.val = pluginsFilterJSON.val;
        }

        public void DeregisterPMEFileNotification(JSONStorableString pmeJSON)
        {
            pmeNotiticationList.Remove(pmeJSON);
        }

        public void RegisterCreatorNameNotification(JSONStorableString creatorNameJSON, JSONStorableBool creatorNameIsCollectionJSON)
        {
            creatorNameNotificationList.Add(creatorNameJSON);
            creatorNameIsCollectionNotificationList.Add(creatorNameIsCollectionJSON);
 //           creatorNameJSON.val = creatorFilterJSON.val;
            creatorNameIsCollectionJSON.val = allCreators || localCreators;
        }

        public void DeregisterCreatorNameNotification(JSONStorableString creatorNameJSON, JSONStorableBool creatorNameIsCollectionJSON)
        {
            creatorNameNotificationList.Remove(creatorNameJSON);
            creatorNameIsCollectionNotificationList.Remove(creatorNameIsCollectionJSON);
        }

        protected override void CreateGroupUI(int leftRight, int groupID)
        {
            PluginAssist pa;
            PluginManifestEntry pme;
            switch (groupID)
            {
                case 1:
                    CreateFilterablePopup(creatorFilterJSON, false, 800, CreatorChoicesRefresh, null, true, 1);
                    if (showConcealedOption) CreateToggle(showConcealedCreatorsJSON);
                    break;
                case 2:
                    CreateSpacer(10,false,2);
                    CreatePopup(pluginTypeFilterJSON, false, 400, PluginTypeFilterChoicesRefresh, null, true, true, false, 2);
                    break;
                case 3:
                    if (PC.patreonContentEnabled) CreatePopup(tagsFilterJSON, false, 400, PluginTagFilterChoicesRefresh, null, true, true, false, 3);
                    break;
                case 4:
                    CreateFilterablePopup(pluginsFilterJSON, false, 400, PluginChoicesRefresh, null, true, 4);
                    break;
                case 5:
                    pa = (PluginAssist)MVRS;
                    pme = pa.pluginManifest.GetPluginManifestEntry(pluginsFilterJSON.val);
                    if (pme!=null && showVARVersions && !pme.isLocal && pme.varVersionCount > 1) CreatePopup(packageVersionsFilterJSON, false, 300, PackageVersionChoicesRefresh, null, true, true, false, 5);
                    break;
                case 6:
                    pa = (PluginAssist)MVRS;
                    pme = pa.pluginManifest.GetPluginManifestEntry(pluginsFilterJSON.val);

                    if (pme != null)
                    {
                        int packageVersion = -1;
                        if (pme.isLocal) packageVersion = 0;
                        else packageVersion = int.Parse(packageVersionsFilterJSON.val);

                        if (packageVersion >= 0 && pathVariantFilterJSON.choices.Count > 1)
                        {
                            CreatePopup(pathVariantFilterJSON, false, 300, PathVariantChoicesRefresh, null, true, true, false, 6);
                        }                    
                        SetPluginFileReference();
                    }
                    if (showConcealedOption) CreateToggle(showConcealedPluginsJSON, false, 50f,6);

                    if (!pa.settings.hideHelpText.val)
                    {
                        float height = Mathf.Max(0f, 840f - GetTotalUIElementHeights(LeftRight.left));
                        if (showConcealedOption) height -= 70f;
                        CreateSpacer(height, false, 6);
                        var textBox = CreateTextField(new JSONStorableString("", HelpText.filterPluginsHelpText), false, false, InputField.LineType.SingleLine, 6);
                        textBox.height = 270f;
                        textBox.backgroundColor = Color.grey;
                        textBox.textColor = Color.white;
                    }

                    break;
            }
        }

        public void RefreshUI(bool activateUI = false, bool _showConcealedOption=false, bool _showVARVersions=false, bool _hideCreatorGroups =false)
        {
            showConcealedOption = _showConcealedOption;
            showVARVersions = _showVARVersions;
            hideCreatorGroups = _hideCreatorGroups;

            if (activateUI) activeUI = true;
            if (activeUI)
            {
                RefreshGroupUI(LeftRight.left, 1);
            }
        }  

        private void CreatorChoicesRefresh()
        {
            PluginAssist pa = (PluginAssist)MVRS;
            bool showConcealed = showConcealedOption && showConcealedCreatorsJSON.val;

            List<string> creatorsList = pa.pluginManifest.GetCreatorsList(showConcealed,PreferredFilter.preferred);
            List<string> notPreferredCreatorsList = pa.pluginManifest.GetCreatorsList(showConcealed, PreferredFilter.notPreferred);
            creatorsList.AddRange(notPreferredCreatorsList);

            List<string> fixedIndexCreators = new List<string>();
            if (!hideCreatorGroups)
            {
                fixedIndexCreators.Add("ALL");
                fixedIndexCreators.Add("VAR ONLY");
                fixedIndexCreators.Add("LOCAL ONLY");
            }
            creatorFilterJSON.SetIndexChoices(fixedIndexCreators,creatorsList);
        }
        private void FixedIndexCreatorSelected(JSONStorableStringChooser creator)
        {
        }
        private void CreatorSelected(JSONStorableStringChooser creator)
        {
        }

        private void FixedIndexTagSelected(string tag)
        {
        }
        private void TagSelected(string tag)
        {
        }

        private void PluginTypeSelected(string val)
        {
            foreach (JSONStorableString jss in pluginTypeFilterNotiticationList) jss.val = pluginTypeFilterJSON.val;
        }

        private void PathVariantSelected(string path)
        {       
            SetPluginFileReference();
        }
        private void CreatorUpdateNotification()
        {
            foreach (JSONStorableBool jsb in creatorNameIsCollectionNotificationList) jsb.val = allCreators ;

            string creatorName = creatorFilterJSON.displayVal;
            if (localCreators) creatorName = "LOCAL";
            foreach (JSONStorableString jss in creatorNameNotificationList) jss.val = creatorName;
        }

        private void PluginTypeFilterChoicesRefresh()
        {
            PluginAssist pa = (PluginAssist)MVRS;
            List<string> choices = pa.pluginManifest.GetPluginTypesList(creatorFilterJSON, true);
            List<string> displayChoices = pa.pluginManifest.GetPluginTypesList(creatorFilterJSON, false);

            choices.Insert(0,"0");
            displayChoices.Insert(0,"All");

            pluginTypeFilterJSON.choices = choices;
            pluginTypeFilterJSON.displayChoices = displayChoices;
        }

        private void PluginTagFilterChoicesRefresh()
        {
            PluginAssist pa = (PluginAssist)MVRS;
            List<string> displayChoices = pa.pluginManifest.GetPluginTagsList(creatorFilterJSON, pluginTypeFilterJSON);
            displayChoices.Sort();
            displayChoices.Insert(0, "All");

            List<string> choices = new List<string>();
            for (int i = 0; i < displayChoices.Count; i++) choices.Add(i.ToString());

            tagsFilterJSON.choices = choices;
            tagsFilterJSON.displayChoices = displayChoices;
        }

        private void PluginChoicesRefresh()
        {
            PluginAssist pa = (PluginAssist)MVRS;

            bool showConcealedPlugins = showConcealedOption && showConcealedPluginsJSON.val;
            bool showConcealedCreators = showConcealedOption && showConcealedCreatorsJSON.val;

            var choices = pa.pluginManifest. GetPluginsList(allCreators, localCreators, creatorFilterJSON.displayVal, false,PreferredFilter.preferred, showConcealedCreators, showConcealedPlugins, int.Parse(pluginTypeFilterJSON.val), tagsFilterJSON);
            var displayChoices = pa.pluginManifest.GetPluginsList(allCreators, localCreators, creatorFilterJSON.displayVal, true, PreferredFilter.preferred, showConcealedCreators, showConcealedPlugins, int.Parse(pluginTypeFilterJSON.val), tagsFilterJSON);

            Utils.SortLists(ref displayChoices, ref choices);
            
            var choicesNP = pa.pluginManifest.GetPluginsList(allCreators, localCreators, creatorFilterJSON.displayVal, false, PreferredFilter.notPreferred, showConcealedCreators, showConcealedPlugins, int.Parse(pluginTypeFilterJSON.val), tagsFilterJSON);
            var displayChoicesNP = pa.pluginManifest.GetPluginsList(allCreators, localCreators, creatorFilterJSON.displayVal, true, PreferredFilter.notPreferred, showConcealedCreators, showConcealedPlugins, int.Parse(pluginTypeFilterJSON.val), tagsFilterJSON);

            Utils.SortLists(ref displayChoicesNP, ref choicesNP);

            choices.AddRange(choicesNP);
            displayChoices.AddRange(displayChoicesNP);

            if (choices.Count<1)
            {
                choicesNP.Add("0");
                displayChoicesNP.Add("No matching plugins available");
            }

            // Do choices first as the way plugin names are displayed is impacted by whether the plugin is preferred or concealed
            pluginsFilterJSON.choices = choices;
            pluginsFilterJSON.displayChoices = displayChoices;
        }

        private void PackageVersionChoicesRefresh()
        {
            PluginAssist pa = (PluginAssist)MVRS;
            PluginManifestEntry pme = pa.pluginManifest.GetPluginManifestEntry(pluginsFilterJSON.val);

            List<string> choices = new List<string>();
            if (pme != null) choices = pme.GetPluginsVersionsList();

            if (choices.Count == 0) choices.Add("No Matching Versions");

            packageVersionsFilterJSON.choices = choices;
        }

        private void PathVariantChoicesRefresh()
        {
            PluginAssist pa = (PluginAssist)MVRS;
            PluginManifestEntry pme = pa.pluginManifest.GetPluginManifestEntry(pluginsFilterJSON.val);

            int packageVersion = -1;
            List<string> choices = new List<string>();
            List<string> displayChoices = new List<string>();
            if (pme != null)
            {
                if (pme.isLocal) packageVersion = 0;
                else if (pme.varVersionCount == 1) packageVersion = pme.highestVARNumber;
                else if (pme.varVersionCount > 1) packageVersion = int.Parse(packageVersionsFilterJSON.val);
                bool showConcealed = showConcealedOption && showConcealedPluginsJSON.val;

                choices = pme.GetPathVariantsList(packageVersion, false, PreferredFilter.preferred, showConcealed);
                displayChoices = pme.GetPathVariantsList(packageVersion, true, PreferredFilter.preferred, showConcealed);

                List<string> choicesNotPreferred = pme.GetPathVariantsList(packageVersion, false, PreferredFilter.notPreferred, showConcealed);
                List<string> displayChoicesNotPreferred = pme.GetPathVariantsList(packageVersion, true, PreferredFilter.notPreferred, showConcealed);

                choices.AddRange(choicesNotPreferred);
                displayChoices.AddRange(displayChoicesNotPreferred);
            }
            if (choices.Count == 0)
            {
                choices.Add("-1");
                displayChoices.Add("No Paths available");
            }
            pathVariantFilterJSON.choices = choices;
            pathVariantFilterJSON.displayChoices = displayChoices;
        }

        private void SetPluginFileReference()
        {
            PluginAssist pa = (PluginAssist)MVRS;
            PluginManifestEntry pme = pa.pluginManifest.GetPluginManifestEntry(pluginsFilterJSON.val);
            CreatorUpdateNotification();
            string path = "";
            if (pme != null)
            {                
                if (pme.isLocal) path = pathVariantFilterJSON.val;
                else
                {
                    int varVersion = int.Parse(packageVersionsFilterJSON.val);
                    if (pme.GetPackageVersionsCount() == 1) varVersion = pme.highestVARNumber;

                    path = pathVariantFilterJSON.val;
                    path = pme.creator + "." + pme.varPackageName + "." + varVersion.ToString() + ":/" + path;
                }
                pluginFileReference = path + "/" + pme.pluginName;
                
                foreach (JSONStorableString jss in pmeNotiticationList) jss.val = pluginsFilterJSON.val;
                foreach (JSONStorableString jss in pluginFileNotiticationList) jss.val = pluginFileReference;
            }
            else Utils.LogDebug("SetPluginFileReference: PME is NULL");
        }
    }

    public class PluginFilterSettingsTabUI : BaseTabUI
    {
        Settings settings;

        public PluginFilterSettingsTabUI(MVRScript mvrs) : base(mvrs)
        {
            PluginAssist pa = (PluginAssist)MVRS;
            settings = pa.settings;
        }
        protected override void CreateGroupUI(int leftRight, int groupID)
        {
            PluginAssist pa;
            pa = (PluginAssist)MVRS;
            switch (groupID)
            {
                case 1:
                    CreateToggle(settings.concealedFilterItemsTagEnabled,false,80f);
                    CreateToggle(settings.preferredFilterItemsTagEnabled, false, 80f);

                    if (!pa.settings.hideHelpText.val)
                    {
                        float height = Mathf.Max(0f, 840f - GetTotalUIElementHeights(LeftRight.left)  - 65f );
                        CreateSpacer(height, false, 1);
                        var textBox = CreateTextField(new JSONStorableString("", HelpText.setPluginFiltersHelpText), false, false, InputField.LineType.SingleLine, 1);
                        textBox.height = 270f;
                        textBox.backgroundColor = Color.grey;
                        textBox.textColor = Color.white;
                    }
                    break;
            }
        }
        public void RefreshUI(bool activateUI = false)
        {
            if (activateUI) activeUI = true;
            if (activeUI)
            {
                RefreshGroupUI(LeftRight.right, 1);
            }
        }
        
    }
    public class GeneralSettingsTabUI : BaseTabUI
    {
        PluginManifest pluginManifest;
        Settings settings;
        public GeneralSettingsTabUI(MVRScript mvrs) : base(mvrs)
        {
            PluginAssist pa = (PluginAssist)MVRS;
            pluginManifest = pa.pluginManifest;
            settings = pa.settings;
            settings.hideHelpText.setCallbackFunction = this.HideHelpTextUpdated;
        }
        protected override void CreateGroupUI(int leftRight, int groupID)
        {
            PluginAssist pa;
            pa = (PluginAssist)MVRS;

            if (leftRight == LeftRight.left)
            {
                switch (groupID)
                {
                    case 1:
                        var button = CreateButton("Refresh VAR Plugin Manifest", false, RefreshVARManifest);
                        button.buttonColor = Color.yellow;
                        var button1 = CreateButton("Refresh Local Plugin Manifest", false, RefreshLocalManifest);
                        button1.buttonColor = Color.yellow;

                        if (!pa.settings.hideHelpText.val)
                        {
                            float height = Mathf.Max(0f, 840f - GetTotalUIElementHeights(LeftRight.left) - 65f);
                            CreateSpacer(height, false, 1);
                            var textBox = CreateTextField(new JSONStorableString("", HelpText.setGeneralLeftHelpText), false, false, InputField.LineType.SingleLine, 1);
                            textBox.height = 270f;
                            textBox.backgroundColor = Color.grey;
                            textBox.textColor = Color.white;
                        }

                        break;
                }
            }
            else if(leftRight == LeftRight.right)
            {
                switch (groupID)
                {
                    case 1:
                        CreateToggle(settings.hideHelpText, true);
                        CreateToggle(settings.showDebugLogging, true);

                        if (!pa.settings.hideHelpText.val)
                        {
                            float height = Mathf.Max(0f, 840f - GetTotalUIElementHeights(LeftRight.right) + 10f - 65f);
                            CreateSpacer(height, true, 1);
                            var textBox = CreateTextField(new JSONStorableString("", HelpText.setGeneralRightHelpText), true, false, InputField.LineType.SingleLine, 1);
                            textBox.height = 270f;
                            textBox.backgroundColor = Color.grey;
                            textBox.textColor = Color.white;
                        }

                        break;
                }
            }
            
        }
        public void HideHelpTextUpdated(bool val)
        {
            RefreshGroupUI(LeftRight.both, 1);
        }

        public void RefreshUI(bool activateUI = false)
        {
            if (activateUI) activeUI = true;
            if (activeUI)
            {
                RefreshGroupUI(LeftRight.both, 1);
            }
        }

        private void RefreshVARManifest()
        {
            SuperController.LogMessage("PluginAssist: VAR Plugin Manifest Refresh started");
            MVRS.StartCoroutine(RefreshVARManifestNextFrame());            
        }
        private void RefreshLocalManifest()
        {
            SuperController.LogMessage("PluginAssist: Local Plugin Manifest Refresh started");
            MVRS.StartCoroutine(RefreshLocalManifestNextFrame());  
        }

        private IEnumerator RefreshVARManifestNextFrame()
        {
            yield return new WaitForSeconds(0.1f);
            pluginManifest.RefreshVARPluginManifest();
            SuperController.LogMessage("PluginAssist: VAR Plugin Manifest Refresh finished");
        }
        private IEnumerator RefreshLocalManifestNextFrame()
        {
            yield return new WaitForSeconds(0.1f);
            pluginManifest.RefreshLocalPluginManifest();
            SuperController.LogMessage("PluginAssist: Local Plugin Manifest Refresh finished");
        }

    }

    public class AboutSettingsTabUI : BaseTabUI
    {
        JSONStorableString pluginDetailsJSON;
        public AboutSettingsTabUI(MVRScript mvrs) : base(mvrs)
        {
            string pluginDetails ;
            if (PC.patreonContentEnabled) pluginDetails = "PluginAssist v" + Const.PAversion + "\nby JayJayWon\nPatron version\n\nhttps://www.patreon.com/JayJayWon"; 
            else pluginDetails = "PluginAssist v" + Const.PAversion + "\nby JayJayWon\nFree version\n\nhttps://www.patreon.com/JayJayWon";

            pluginDetailsJSON = new JSONStorableString("PluginDetails", pluginDetails);
        }
        protected override void CreateGroupUI(int leftRight, int groupID)
        {
            switch (groupID)
            {
               case 1:
                    var textBox = CreateTextField(pluginDetailsJSON);
                    textBox.backgroundColor = Color.grey;
                    textBox.textColor = Color.cyan;
                    break;
            }
        }
        public void RefreshUI(bool activateUI = false)
        {
            if (activateUI) activeUI = true;
            if (activeUI)
            {
                RefreshGroupUI(LeftRight.left, 1);
            }
        }

    }

    public class ModifyTagsCategoriesTabUI : BaseTabUI
    {
        TagsManifest tagManifest;
        private JSONStorableStringChooser tagSelectorJSON;
        private JSONStorableString tagLabelJSON;

        private PluginTag currentPluginTag;

        private bool tagDeletionMode = false;

        public ModifyTagsCategoriesTabUI(MVRScript mvrs) : base(mvrs)
        {
            PluginAssist pa = (PluginAssist)MVRS;
            tagManifest = pa.tagsManifest;
            tagSelectorJSON = new JSONStorableStringChooser("tagSelector", null, "", "Select Tag",TagSelected);
            if (PC.patreonContentEnabled)
            {
                RegisterJSSCRefreshChoicesPreUIRefresh(tagSelectorJSON, PluginTagSelectChoicesRefresh, LeftRight.left, 2);
                RegisterJSSCRefreshUIonValChange(tagSelectorJSON, LeftRight.left, 3);
            }               

            tagLabelJSON = new JSONStorableString("tagLabel", "",TagLabelUpdated);

            uiGroupCount = 4;
        }
        protected override void CreateGroupUI(int leftRight, int groupID)
        {
            PluginAssist pa;
            pa = (PluginAssist)MVRS;
            switch (groupID)
            {
                case 1:
                    if (PC.patreonContentEnabled)
                    {
                        var button = CreateButton("Create new Plugin Tag", false, CreatePluginTag, 1);
                        button.buttonColor = Color.green;
                    }
                    else
                    {
                        var textBox = CreateTextField(new JSONStorableString("", HelpText.tagsNonePatronText), false, false, InputField.LineType.SingleLine, 1);
                        textBox.height = 500f;
                        textBox.backgroundColor = Color.grey;
                        textBox.textColor = Color.cyan;
                    }

                    break;
                case 2:
                    if (tagManifest.tagCount > 0 && PC.patreonContentEnabled)
                    {
                        CreatePopup(tagSelectorJSON, false, 600f, PluginTagSelectChoicesRefresh, null, true, true, true, 2);
                    }
                    
                    break;
                case 3:
                    if (tagManifest.tagCount > 0 && PC.patreonContentEnabled)
                    {
                        CreateTextField(tagLabelJSON, false, true, InputField.LineType.MultiLineSubmit, 3);
                        currentPluginTag = tagManifest.GetTag(tagSelectorJSON.val);
                        tagLabelJSON.valNoCallback = currentPluginTag.tagLabel;

                        UIDynamicButton button2;

                        if (!tagDeletionMode)
                        {
                            button2 = CreateButton("Permanently Delete Plugin Tag...", false, StartTagDeletionMode, 3);
                            button2.buttonColor = Color.red;
                            button2.buttonText.color = Color.white;
                        }
                        else
                        {
                            button2 = CreateButton("Cancel Plugin Tag Deletion", false, CancelTagDeletionMode, 3);
                            button2.buttonColor = Color.yellow;
                            button2.buttonText.color = Color.black;
                        }
                    }
                    break;
                case 4:
                    if (tagDeletionMode && PC.patreonContentEnabled)
                    {
                        var button3 = CreateButton("Confirm PERMANENT Tag Deletion", false, DeletePluginTag, 4);
                        button3.buttonColor = Color.red;
                        button3.buttonText.color = Color.white;
                    }
                    if (!pa.settings.hideHelpText.val && PC.patreonContentEnabled)
                    {
                        float height = Mathf.Max(0f, 840f - GetTotalUIElementHeights(LeftRight.left) + 5f - 65f -100f);
                        CreateSpacer(height, false, 4);
                        var textBox = CreateTextField(new JSONStorableString("", HelpText.catPluginTagsHelpText), false, false, InputField.LineType.SingleLine, 4);
                        textBox.height = 270f;
                        textBox.backgroundColor = Color.grey;
                        textBox.textColor = Color.white;
                    }

                    break;
            }
        }
        public void RefreshUI(bool activateUI = false)
        {
            if (activateUI)
            {
                activeUI = true;
                tagDeletionMode = false;
            }
            if (activeUI)
            {
                RefreshGroupUI(LeftRight.left, 1);
            }
        }

        private void TagLabelUpdated(string tagLabel)
        {
            tagDeletionMode = false;
            if (tagLabel != currentPluginTag.tagLabel)
            {
                string newTagLabel = tagManifest.TryLabelSet(currentPluginTag, tagLabel);
                tagSelectorJSON.val = newTagLabel;
            }
            
        }

        private void CreatePluginTag()
        {
            var pluginTag = tagManifest.NewTag();
            tagSelectorJSON.val = pluginTag.tagLabel;
            tagDeletionMode = false;
            RefreshGroupUI(LeftRight.left, 1);
        }

        private void StartTagDeletionMode()
        {
            tagDeletionMode = true;
            RefreshGroupUI(LeftRight.left, 2);
        }

        private void CancelTagDeletionMode()
        {
            tagDeletionMode = false;
            RefreshGroupUI(LeftRight.left, 2);
        }

        private void DeletePluginTag()
        {
            if (currentPluginTag != null) tagManifest.DeleteTag(currentPluginTag);

            tagDeletionMode = false;
            RefreshGroupUI(LeftRight.left, 2);
        }

        private void TagSelected(string tagLabel)
        {
            tagDeletionMode = false;
        }

        private void PluginTagSelectChoicesRefresh()
        {
            List<string> choices = tagManifest.GetTagLabels();
            choices.Sort();
            tagSelectorJSON.choices = choices;
        }
    }

}
